package tiny

import java.io.{File, FileReader}

/**
  * A Tiny Calculator Language Compiler
  *
  * Our compiler translates calculator source code to x86 machine code
  * in stages using multiple internal and external program representations.
  *
  * External program representations:
  * - Tiny source code
  * - x86 assembly language code
  * - x86 machine code
  *
  * Internal program representations:
  * - AST.scala
  * AST: Abstract Syntax Tree
  * - IR.scala
  * IR: Intermediate Representation/Three-Address Code
  *
  * Compilation stages:
  * - Parse.scala
  * Lexing and Parsing: source to AST
  * - ScopeCheck.scala:
  * Scope checking
  * - IRGen.scala
  * IR Generation: AST to IR
  * - Opt.scala
  * Optimization: IR to IR
  * - CodeGen.scala
  * Code Generation: IR to x86 assembly
  * - Link.scala
  * [External] Assembly and Linking: assembly to machine code executable
  *
  */
object Compiler {
  def main(args: Array[String]): Unit = {
    try {
      val source = new Lexer(new FileReader(new File(args(0))))

      // 0. PARSE
      // Parse the source code character sequence to an Abstract Syntax Tree.
      println("== AST ===========")
      val ast: Program = new parser(source).parse().value match { case p: Program => p }
      println(ast)

      // 1. CHECK
      // Check variable uses to ensure they follow definitions respecting scope.
      println("== CHECK =========")
      ScopeCheck(ast)

      // 2. IR
      // Convert the AST to a Three-Address Code Intermediate Representation.
      println("== IR ===========")
      val ir = IRGen(ast)
      for (inst <- ir) println("  " + inst)

      // 3. OPTIMIZE
      // Optimize the IR program.
      println("== OPT ===========")
      val irOpt = Opt(ir)

      // 4. CODEGEN
      // Translate the TAC into x86 assembly code.
      println("== ASM ===========")
      val asm = CodeGen(irOpt)
      println(asm)

      // 5. LINK
      // [External] Assemble the generated code and link a binary executable.
      println("== BIN ===========")
      if (args.isEmpty) Link(asm, "stdin.s", "stdin.bin")
      else Link(asm, args(0) + ".s", args(0) + ".bin")

      println("================")
      println("ACCEPTED")
    }
    // Commented this out to see where spurious errors originate.
    catch {
      // Die with compiler errors.
      case e: CompilerError => {
        println("Tiny " + e.getClass.getSimpleName + ": " + e.getMessage)
        println("================")
        println("REJECTED")
      }
    }
  }
}

/** Common basis for various types of compiler errors. */
class CompilerError(msg: String) extends Exception(msg)

